/* Copyright (C) 2002 Edson Pereira, PU1JTE, N1VTN, 7N4NCL
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, download a copy
 * from http://www.fsf.org
 *
 *
 * Propagator algorithm derived from John A. Magliacane, 
 * KD2BD's Predict.
 *
 */

#include <stdio.h>
#include <math.h>
#include <time.h>
#include <sys/timeb.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#include "keppropagator.h"

KepPropagator::KepPropagator()
{
   TP=6.283185307179586;
   PT=1.570796326794897;
   S1=0.397818675;
   C1=0.91746406;
   deg2rad=1.74532925199e-02;
   R0=6378.135;
   FF=3.35289186924e-03;
   KM=1.609344;
}

KepPropagator::~KepPropagator()
{

}

double KepPropagator::FixAngle(double x)
{
    /* This function reduces angles greater than
     two pi by subtracting two pi from the angle */

   while (x>TP)
       x-=TP;

   return x;
}

char *KepPropagator::SubString(char *string, char start, char end)
{
    /* This function returns a substring based on the starting
     and ending positions provided.  It is used heavily in
     the AutoUpdate function when parsing 2-line element data. */

   int x, y;

   if (end >= start)
   {
      for (x = start, y = 0; x <= end && string[x] != 0; x++)
	  if (string[x]!=' ')
	  {
	     temp[y]=string[x];
	     y++;
	  }

      temp[y]=0;
      return temp;
   }
   else
       return NULL;
}

void KepPropagator::CopyString(unsigned char *source,
			    unsigned char *destination,
			    unsigned char start,
			    unsigned char end)
{
    /* This function copies elements of the string "source"
     bounded by "start" and "end" into the string "destination". */

   unsigned j, k=0;

   for (j=start; j<=end; j++)
       if (source[k]!=0)
       {
	  destination[j]=source[k];
	  k++;
       }
}

char KepPropagator::KepCheck(char *line1, char *line2)
{
    /* This function scans line 1 and line 2 of a NASA 2-Line element
     set and returns a 1 if the element set appears to be valid or
     a 0 if it does not.  If the data survives this torture test,
     it's a pretty safe bet we're looking at a valid 2-line
     element set and not just some random garbage that might pass
     as orbital data based on a simple checksum calculation alone. */

   int x;
   unsigned sum1, sum2;

   /* Set up translation table for computing TLE checksums */

   for (x=0; x<=255; val[x]=0, x++);
   for (x='0'; x<='9'; val[x]=x-'0', x++);
   val['-']=1;

    /* Compute checksum for each line */

   for (x=0, sum1=0, sum2=0; x<=67; sum1+=val[(int)line1[x]], sum2+=val[(int)line2[x]], x++);

   /* Perform a "torture test" on the data */   
   x=(val[(int)line1[68]]^(sum1%10)) | (val[(int)line2[68]]^(sum2%10)) |
     (line1[0]^'1')  | (line1[1]^' ')  | (line1[7]^'U')  |
     (line1[8]^' ')  | (line1[17]^' ') | (line1[23]^'.') |
     (line1[32]^' ') | (line1[34]^'.') | (line1[43]^' ') |
     (line1[52]^' ') | (line1[61]^' ') | (line1[63]^' ') |
     (line2[0]^'2')  | (line2[1]^' ')  | (line2[7]^' ')  |
     (line2[11]^'.') | (line2[16]^' ') | (line2[20]^'.') |
     (line2[25]^' ') | (line2[33]^' ') | (line2[37]^'.') |
     (line2[42]^' ') | (line2[46]^'.') | (line2[51]^' ') |
     (line2[54]^'.') | (line1[2]^line2[2]) | (line1[3]^line2[3]) |
     (line1[4]^line2[4]) | (line1[5]^line2[5]) | (line1[6]^line2[6]) |
     (isdigit(line1[68]) ? 0 : 1) | (isdigit(line2[68]) ? 0 : 1) |
     (isdigit(line1[18]) ? 0 : 1) | (isdigit(line1[19]) ? 0 : 1) |
     (isdigit(line2[31]) ? 0 : 1) | (isdigit(line2[32]) ? 0 : 1);
   
   return (x ? 0 : 1);
}

int KepPropagator::init(char* _name, char* line1, char* line2,
		     double _stnlat, double _stnlong, int _stnalt)
{
   int x=0, y;

   stnlat = _stnlat;
   stnlong = _stnlong;
   stnalt = _stnalt;

   strcpy( name, _name );
   strncpy( TLEline1, line1, 80 );
   strncpy( TLEline2, line2, 80 );

   /* Read element set */
   if (KepCheck(line1,line2))
   {
      /* We found a valid TLE! */
      /* Add decimal to eccentricity value */
      line2[25]='.';

      /* Some TLE sources left justify the sat
       name in a 24-byte field that is padded
       with blanks.  The following lines cut
       out the blanks as well as the line feed
       character read by the fgets() function. */

      y=strlen(name);

      while (name[y]==32 || name[y]==0 || name[y]==10 || name[y]==13 || y==0)
      {
	 name[y]=0;
	 y--;
      }

      /* Copy TLE data into the sat data structure */

      strncpy(name,name,24);
      catnum=atol(SubString(line1,2,6));
      year=atoi(SubString(line1,18,19));
      refepoch=atof(SubString(line1,20,31));
      incl=atof(SubString(line2,8,15));
      raan=atof(SubString(line2,17,24));
      eccn=atof(SubString(line2,25,32));
      argper=atof(SubString(line2,34,41));
      meanan=atof(SubString(line2,43,50));
      meanmo=atof(SubString(line2,52,62));
      drag=atof(SubString(line1,33,42));
      orbitnum=atol(SubString(line2,63,67));
      x++;
      return 0;
   }
   else return 1;
}

long KepPropagator::DayNum(int m, int d, int y)
{
    /* This function calculates the day number from m/d/y. */

   long dn;
   double mm, yy;

   if (m<3)
   {
      y--;
      m+=12;
   }

    /* Correct for Y2K... */

   if (y<=50)
       y+=100;

   yy=(double)y;
   mm=(double)m;
   dn=(long)(floor(365.25*(yy-80.0))-floor(19.0+yy/100.0)+floor(4.75+yy/400.0)-16.0);
   dn+=d+30*m+(long)floor(0.6*mm-0.3);
   return dn;
}

double KepPropagator::CurrentDaynum()
{
    /* Read the system clock and return the number
     of days since 31Dec79 00:00:00 UTC (daynum 0) */

   struct timeb tptr;
   int x;

   x=ftime(&tptr);

   return ((((double)tptr.time+0.001*(double)tptr.millitm)/86400.0)-3651.0);
}

int KepPropagator::daynum2unix(double daynum)
{
   return (int)rint(86400.0*(daynum+3651.0));
}

char *KepPropagator::daynum2String(double daynum)
{
/* This function takes the given epoch as a fractional number of
    days since 31Dec79 00:00:00 UTC and returns the corresponding
    date as a string of the form "Tue 12Oct99 17:22:37". */

   static char timestr[25];
   time_t t;

   /* Convert daynum to Unix time (seconds since 01-Jan-70) */
   t=(time_t)rint(86400.0*(daynum+3651.0));

   sprintf(timestr,"%s",asctime(gmtime(&t)));
   timestr[ strlen( timestr ) - 1 ] = '\0';
   return timestr;
}

double KepPropagator::PreCalc( double daynum )
{
   /* This function performs preliminary calculations
    prior to tracking or prediction. */

   epoch=DayNum(1,0,year)+refepoch;
   age=daynum-epoch;
   yr=(float)year;

   /* Do the Y2K thing... */

   if (yr<=50.0)
       yr+=100.0;

   t1=yr-1.0;
   df=366.0+floor(365.25*(t1-80.0))-floor(t1/100.0)+floor(t1/400.0+0.75);
   t1=(df+29218.5)/36525.0;
   t1=6.6460656+t1*(2400.051262+t1*2.581e-5);
   se=t1/24.0-yr;
   n0=meanmo+(age*drag);
   sma=331.25*pow((1440.0/n0),(2.0/3.0));
   e2=1.0-(eccn*eccn);
   e1=sqrt(e2);
   k2=9.95*(exp(log(R0/sma)*3.5))/e2*e2;
   s1=sin(incl*deg2rad);
   c1=cos(incl*deg2rad);
   l8=stnlat*deg2rad;
   s9=sin(l8);
   c9=cos(l8);
   s8=sin(-stnlong*deg2rad);
   c8=cos(stnlong*deg2rad);
   r9=R0*(1.0+(FF/2.0)*(cos(2.0*l8)-1.0))+(stnalt/1000.0);
   l8=atan((1.0-FF)*(1.0-FF)*s9/c9);
   z9=r9*sin(l8);
   x9=r9*cos(l8)*c8;
   y9=r9*cos(l8)*s8;
   apogee=sma*(1.0+eccn)-R0;
   perigee=sma*(1.0-eccn)-R0;

   return daynum;
}

void KepPropagator::Calc()
{
   /* This is the stuff we need to do repetitively. */

   age=Daynum-epoch;
   o=deg2rad*(raan-(age)*k2*c1);
   s0=sin(o);
   c0=cos(o);
   w=deg2rad*(argper+(age)*k2*(2.5*c1*c1-0.5));
   s2=sin(w);
   c2=cos(w);
   c[1][1]=c2*c0-s2*s0*c1;
   c[1][2]=-s2*c0-c2*s0*c1;
   c[2][1]=c2*s0+s2*c0*c1;
   c[2][2]=-s2*s0+c2*c0*c1;
   c[3][1]=s2*s1;
   c[3][2]=c2*s1;
   q0=(meanan/360.0)+orbitnum;
   q=n0*age+q0;
   rv=(long)floor(q);
   q=q-floor(q);
   m=q*TP;
   e=m+eccn*(sin(m)+0.5*eccn*sin(m*2.0));

   do   /* Kepler's Equation */
   {
      s3=sin(e);
      c3=cos(e);
      r3=1.0-eccn*c3;
      m1=e-eccn*s3;
      m5=m1-m;
      e=e-m5/r3;

   }
   while (fabs(m5)>=1.0e-6);

   x0=sma*(c3-eccn);
   yzero=sma*e1*s3;
   r=sma*r3;
   x1=x0*c[1][1]+yzero*c[1][2];
   yone=x0*c[2][1]+yzero*c[2][2];
   z1=x0*c[3][1]+yzero*c[3][2];
   g7=(Daynum-df)*1.0027379093+se;
   g7=TP*(g7-floor(g7));
   s7=-sin(g7);
   c7=cos(g7);
   x=x1*c7-yone*s7;
   y=x1*s7+yone*c7;
   z=z1;
   x5=x-x9;
   y5=y-y9;
   z5=z-z9;
   range=x5*x5+y5*y5+z5*z5;
   z8=x5*c8*c9+y5*s8*c9+z5*s9;
   x8=-x5*c8*s9-y5*s8*s9+z5*c9;
   y8=y5*c8-x5*s8;
   ak=r-R0;
   elevation=atan(z8/sqrt(range-z8*z8))/deg2rad;
   azimuth=atan(y8/x8)/deg2rad;

   if (x8<0.0)
       azimuth+=180.0;

   if (azimuth<0.0)
       azimuth+=360.0;

   ma256=(int)(256.0*q);
   am=ak/KM;
   rk=sqrt(range);
   rm=rk/KM;
   vk=3.6*sqrt(3.98652e+14*((2.0/(r*1000.0))-1.0/(sma*1000.0)));
   vm=vk/KM;
   fk=12756.33*acos(R0/r);
   fm=fk/KM;
   ssplat=atan(z/sqrt(r*r-z*z))/deg2rad;
   ssplong=-atan(y/x)/deg2rad;

   if (x<0.0)
       ssplong+=180.0;

   if (ssplong<0.0)
       ssplong+=360.0;

   irk=(long)rint(rk);
   isplat=(int)rint(ssplat);
   isplong=(int)rint(ssplong);
   iaz=(int)rint(azimuth);
   iel=(int)rint(elevation);
}

char KepPropagator::AosHappens()
{
    /* This function returns a 1 if the satellite can ever rise
     above the horizon of the ground station. */

   double lin, sma, apogee;

   lin=incl;

   if (lin >= 90.0)
       lin=180.0-lin;

   sma=331.25*exp(log(1440.0/meanmo)*(2.0/3.0));
   apogee=sma*(1.0+eccn)-R0;

   if ((acos(R0/(apogee+R0))+(lin*deg2rad)) > fabs(stnlat*deg2rad))
       return 1;
   else
       return 0;
}

char KepPropagator::Decayed()
{
    /* This function returns a 1 if it appears that the
     satellite has decayed at the time of 'daynum' */

   double satepoch;

   satepoch=DayNum(1,0,year)+refepoch;

   if (satepoch+((16.666666-meanmo)/(10.0*fabs(drag))) < Daynum)
       return 1;
   else
       return 0;
}

char KepPropagator::Geostationary()
{
    /* This function returns a 1 if the satellite appears
     to be in a geostationary orbit */

   if (fabs(meanmo-1.0027)<0.0002)

       return 1;
   else
       return 0;
}

void KepPropagator::propagate( double _daynum )
{
   Daynum = _daynum;
   PreCalc( Daynum );
   Calc();

   // Save state
   Catnum = catnum;
   SSPlat = ssplat;
   SSPlon = ssplong;
   Azimuth = azimuth;
   Elevation = elevation;
   Altitude = ak;
   SlantRange = rk;
   Velocity = vk;
   FootPrint = fk;
   Phase = q;
   OrbitNumber = rv;
   
   dt = Daynum * 86400.0 - old_clock;
   dp = ( SlantRange * 1000.0 ) - old_range;
   old_clock = 86400.0 * Daynum;
   old_range = SlantRange * 1000.0;

}

char *KepPropagator::getSatName()
{
   static char satName[20];
   sprintf( satName, name );
   return satName;
}

long KepPropagator::getCatalogNumber()
{
   return Catnum;
}

double KepPropagator::getDaynum()
{
   return Daynum;
}

double KepPropagator::getSspLat()
{
   return SSPlat;
}

double KepPropagator::getSspLon()
{
   return SSPlon;
}

double KepPropagator::getAzimuth()
{
   return Azimuth;
}

double KepPropagator::getElevation()
{
   return Elevation;
}

double KepPropagator::getAltitude()
{
   return Altitude;
}

double KepPropagator::getSlantRange()
{
   return SlantRange;
}

double KepPropagator::getVelocity()
{
   return Velocity;
}

double KepPropagator::getFootprint()
{
   return FootPrint;
}

double KepPropagator::getPhase()
{
   return Phase;
}

long KepPropagator::getOrbitNumber()
{
   return OrbitNumber;
}

double KepPropagator::getDoppler( double f )
{
   doppler = 0.0;
   if ( dt > 0.0 ) doppler = -(( dp/dt ) / 299792458.0 ) * f;
   return doppler;
}

double KepPropagator::getPathLoss( double f )
{
   // PathLoss = 20 * log10(SlantRange * 4PI * Freq/Speed of light) 
   return 20 * log10( rk * 4 * M_PI * f / 1.0e6 * 299.792458 );  
}

char *KepPropagator::getTLELine1()
{
   return TLEline1;
}

char *KepPropagator::getTLELine2()
{
   return TLEline2;
}
